//******************************************************************************
//  LifeCanvas.java:    LifeCanvas class
//      (c) 1997 Josef Pelikan, KSVI MFF UK Praha
//          Josef.Pelikan@mff.cuni.cz, http://sun3.ms.mff.cuni.cz/~pepca/
//******************************************************************************

import java.awt.*;

//=======================================================================
//  LifeCanvas: rectangle array of cells
//=======================================================================

class LifeCanvas extends Canvas
{
        // playfield:
    protected LifeEngine engine;    // playfield representation, next generation
    protected int maxAge;           // maximal cell age (for color arrays only)

        // cell rendering:
    protected Dimension preferred;  // preferred canvas size in pixels
    protected int cellSize;         // cell size in pixels
    protected int x0, y0;           // rendering origin

    protected int myWidth, myHeight; // canvas size

    protected Color[] colors;       // cell colors (0 .. maxAge)
    protected Image[] imageCache;   // cell image cache (0 .. maxAge)

    protected boolean drawBoxGrid;  // box grid
    protected static Color gridColor = new Color(192,192,192);

    protected boolean paintAll = true; // repaint the whole canvas?
    protected boolean toPaint = false; // painting request

    protected int repaintCycle = 1; // repainting cycle for fast animation

    public final static int ciRGB = 0;      // RGB color interpolation
    public final static int ciHSBinc = 1;   // HSB color interpolation (increment)
    public final static int ciHSBdec = -1;  // HSB color interpolation (decrement)

    public LifeCanvas ( Dimension preferred, int cellSize, int maxAge, int colorInt,
                        Color young, Color old )
            // preferred size, cell size, maximum cell age (>0),
            // type of color interpolation (0 .. RGB, +-1 .. HSB), extremal colors
    {
            // determine dimensions:
        this.preferred = preferred;
        this.cellSize = cellSize;
        x0 = y0 = 0;
            // playfield:
        this.maxAge = (maxAge < 1) ? 1 : (maxAge > 255) ? 255 : maxAge;
        LifeBox.maxAge = this.maxAge;
        engine = new LifeEngine();
            // color scale:
        colors = new Color[this.maxAge+1];
        colors[0] = Color.black;    // background
        colors[this.maxAge] = old;  // extremal colors
        colors[1] = young;
        if ( colorInt == ciRGB ) {      // RGB interpolation
            float R = young.getRed()/255.0f;
            float dR = (old.getRed()/255.0f - R) / (this.maxAge-1);
            float G = young.getGreen()/255.0f;
            float dG = (old.getGreen()/255.0f - G) / (this.maxAge-1);
            float B = young.getBlue()/255.0f;
            float dB = (old.getBlue()/255.0f - B) / (this.maxAge-1);
                // color interpolation:
            for ( int c = 2; c < this.maxAge; c++ ) {
                R += dR; G += dG; B += dB;
                colors[c] = new Color(R,G,B);
                }
            }
        else {                          // HSB interpolation
            float HSB[] = Color.RGBtoHSB(young.getRed(),young.getGreen(),young.getBlue(),null);
            float dHSB[] = Color.RGBtoHSB(old.getRed(),old.getGreen(),old.getBlue(),null);
            if ( colorInt > ciRGB ) {
                if ( dHSB[0] < HSB[0] ) dHSB[0] += 1.0f;
                }
            else
                if ( dHSB[0] > HSB[0] ) dHSB[0] -= 1.0f;
            for ( int i = 0; i < 3; i++ )   // compute differences:
                dHSB[i] = (dHSB[i] - HSB[i]) / (this.maxAge-1);
                // color interpolation:
            for ( int c = 2; c < this.maxAge; c++ ) {
                for ( int i = 0; i < 3; i++ ) HSB[i] += dHSB[i];
                if ( HSB[0] < 0.0 ) HSB[0] += 1.0f;
                if ( HSB[0] > 1.0 ) HSB[0] -= 1.0f;
                colors[c] = Color.getHSBColor(HSB[0],HSB[1],HSB[2]);
                }
            }
    }   // LifeCanvas.LifeCanvas()


    public void fixSize ()
    {
        Rectangle myBounds = bounds();
        myWidth = myBounds.width;
        myHeight = myBounds.height;
    }   // LifeCanvas.fixSize()


    private void createImageCache ()
    {
        imageCache = new Image[maxAge+1];
        int border = cellSize / 4;
        int xPoints[] = { border, cellSize-border, cellSize, cellSize, cellSize-border-1, border+1, 0, 0 };
        int yPoints[] = { 0, 0, border, cellSize-border-1, cellSize, cellSize, cellSize-border-1, border };
        for ( int c = 0; c <= maxAge; c++ ) {
            imageCache[c] = createImage(cellSize,cellSize);
            Graphics g = imageCache[c].getGraphics();
            g.setColor(colors[0]);
            g.fillRect(0,0,cellSize,cellSize);
            g.setColor(colors[c]);
            g.fillPolygon(xPoints,yPoints,8);
            }
    }   // LifeCanvas.createImageCache()


    public Dimension minimumSize ()
    {
        return new Dimension(144,48);
    }   // LifeCanvas.minimumSize()


    public Dimension preferredSize ()
    {
        return preferred;
    }   // LifeCanvas.preferredSize()


    public boolean setBoxGrid ( boolean set )           // sets rendering of box grid
    {
        boolean old = drawBoxGrid;
        drawBoxGrid = set;
        return old;
    }   // LifeCanvas.setBoxGrid()


    private long delay, waiting;        // for redraw
    //private long waitingSum = 0L;

    public void setTimes ( long delay, long waiting )   // sets time labels for rendering
    {
        this.delay = delay;
        this.waiting = waiting;
            // profiling:
        //waitingSum += waiting;
        //System.out.println(""+generation+": ("+delay+") "+waiting+","+waitingSum);
    }   // LifeCanvas.setTimes()


        // color scale rendering:
    final int scaleCellWidth = 4;
    final int scaleCellHeight = 7;
    final int XOrigin = 10;         // used also for stat strings
    final int YOrigin = 10;


    protected byte[] buffer = null;     // label buffer
    protected final static int bufferLen = 256;

    private int printInt ( int pos, int n )                 // returns the final position + 1
    {
        if ( n < 0 ) {
            buffer[pos++] = (byte)'-';
            n = -n;
            }
        int power = 10;
        while ( power <= n ) power *= 10;
        do {
            power /= 10;
            buffer[pos++] = (byte)('0'+(n/power));
            n = n % power;
            } while ( power > 1 );
        return pos;
    }   // LifeCanvas.printInt()


    private void finishDraw ( Graphics g, boolean scale )   // draws color scale, grid, labels
    {
        if ( scale ) {
                // grid:
            if ( drawBoxGrid ) {        // box grid
                g.setColor(gridColor);
                int boxSize = LifeBox.boxSize*cellSize;
                for ( int y = 1 - (y0 % LifeBox.boxSize)*cellSize; y < myHeight; y += boxSize )
                    g.drawLine(1,y,myWidth-1,y);
                for ( int x = 1 - (x0 % LifeBox.boxSize)*cellSize; x < myWidth; x += boxSize )
                    g.drawLine(x,1,x,myHeight-1);
                }
                // color scale:
            for ( int c = 1; c <= LifeBox.maxAge; c++ ) {
                g.setColor(colors[c]);
                g.fillRect(c*scaleCellWidth+XOrigin-(scaleCellWidth-1),YOrigin+1,
                           scaleCellWidth-1,scaleCellHeight-1);
                g.setColor(Color.white);
                g.drawRect(c*scaleCellWidth+XOrigin-scaleCellWidth,YOrigin,
                           scaleCellWidth,scaleCellHeight);
                }
            }
            // labels: generation, number of cells (active cells), waiting time (delay)
        if ( buffer == null ) buffer = new byte[bufferLen];
        g.setColor(colors[0]);
        FontMetrics fm = g.getFontMetrics();
        int height = fm.getHeight();
        int yPos = myHeight-(YOrigin-1)-height;
            // generation:
        int ptr = printInt(0,engine.generation);
        g.fillRect(XOrigin-2,yPos,fm.bytesWidth(buffer,0,ptr)+3,height+2);
            // cells - active cells:
        int cel = ptr;
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.cellNumber);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'-';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.activeNumber);
        int celWid = fm.bytesWidth(buffer,cel,ptr-cel)+3;
        int celPos = myWidth-celWid-(XOrigin-3);
        g.fillRect(celPos-1,yPos,celWid,height+2);
            // box statistics:
        int box = ptr;
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.blinkerBoxes);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'-';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.activeBoxes-engine.blinkerBoxes);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'-';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.nonemptyBoxes);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'-';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.usedBoxes);
        buffer[ptr++] = (byte)'(';
        ptr = printInt(ptr,LifeBox.allocated);
        buffer[ptr++] = (byte)')';
        buffer[ptr++] = (byte)' ';
        int boxWid = fm.bytesWidth(buffer,box,ptr-box)+3;
        int boxPos = (myWidth-boxWid) >> 1;
        g.fillRect(boxPos-1,yPos,boxWid,height+2);
            // timings:
        int tim = ptr;
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,(int)waiting);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'-';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,(int)delay);
        int timWid = fm.bytesWidth(buffer,tim,ptr-tim)+3;
        int timPos = myWidth-timWid-(XOrigin-3);
        g.fillRect(timPos-1,YOrigin-3,timWid,height+2);
            // playfield size:
        int siz = ptr;
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.fieldWidth);
        buffer[ptr++] = (byte)' ';
        buffer[ptr++] = (byte)'x';
        buffer[ptr++] = (byte)' ';
        ptr = printInt(ptr,engine.fieldHeight);
        int sizWid = fm.bytesWidth(buffer,siz,ptr-siz)+3;
        int sizPos = (myWidth-sizWid) >> 1;
        g.fillRect(sizPos-1,YOrigin-3,sizWid,height+2);
            // draw strings:
        g.setColor(Color.white);
        g.drawBytes(buffer,0,cel,XOrigin-1,myHeight-YOrigin);
        g.drawBytes(buffer,cel,box-cel,celPos,myHeight-YOrigin);
        g.drawBytes(buffer,box,tim-box,boxPos,myHeight-YOrigin);
        g.drawBytes(buffer,tim,siz-tim,timPos,(YOrigin-4)+height);
        g.drawBytes(buffer,siz,ptr-siz,sizPos,(YOrigin-4)+height);
            // redraw flag:
        engine.toDraw = false;
        synchronized ( this ) {
			toPaint = false;
			}
    }   // LifeCanvas.finishDraw()


    public void update ( Graphics g )                   // redraws changed cells only
    {
        synchronized ( engine ) {
            if ( paintAll ) paint(g);
            else {
                    // clipping:
                g.clipRect(1,1,myWidth-2,myHeight-2);
                if ( engine.toDraw ) {
                        // cell image cache:
                    if ( imageCache == null ) createImageCache();
                        // draw only changed boxes within clipping region:
                    LifeBox box;
                    int boxSize = LifeBox.boxSize*cellSize;
                        // loop over all changed boxes:
                    while ( (box = engine.getBox()) != null )
                        if ( box.toDraw ) {
                            int origX = (box.x0 - x0)*cellSize;
                            int origY = (box.y0 - y0)*cellSize;
                            if ( origX + boxSize > 1 && origX < myWidth-1 &&
                                 origY + boxSize > 1 && origY < myHeight-1 ) {
                                int y1 = Math.max(0,y0-box.y0);
                                int y2 = Math.min(box.boxSize,y0+1+((myHeight-3)/cellSize)-box.y0);
                                int x1 = Math.max(0,x0-box.x0);
                                int x2 = Math.min(box.boxSize,x0+1+((myWidth-3)/cellSize)-box.x0);
                                int shift0 = box.wordSize - box.boxSize + x1;
                                int ix0 = origX + 1 + x1*cellSize;
                                int iy = origY + 1 + y1*cellSize;
                                for ( int y = y1; y < y2; y++, iy += cellSize ) {
                                    int mask = box.draw[y] << shift0;
                                    box.draw[y] = 0;
                                    if ( LifeBox.maxAge > 1 ) {
                                        for ( int x = x1, ix = ix0; mask != 0 && x < x2; x++, mask += mask, ix += cellSize )
                                            if ( mask < 0 )
                                                g.drawImage(imageCache[ box.ages[y][x] ],ix,iy,this);
                                        }
                                    else {
                                        int cells =  box.cells[y] << shift0;
                                        for ( int x = x1, ix = ix0; mask != 0 && x < x2; x++, mask += mask, cells += cells, ix += cellSize )
                                            if ( mask < 0 )
                                                g.drawImage(imageCache[ (cells<0) ? 1 : 0 ],ix,iy,this);
                                        }
                                    }
                                }
                            box.toDraw = false;
                            }
                    }
                    // draw color palette & stat strings:
                finishDraw(g, (engine.generation & 7) == 0 );
                }
            }
    }   // LifeCanvas.update()


    private void debugFrame ( Graphics g )              // for debugging purposes
    {
        int right = myWidth-1;
        int right4 = right-4;
        int bottom = myHeight-1;
        int bottom4 = bottom-4;

        g.setColor(Color.red);
            // horiz/vert:
        g.drawLine(4,0,4,bottom);
        g.drawLine(0,4,right,4);
        g.drawLine(0,bottom4,right,bottom4);
        g.drawLine(right4,0,right4,bottom);
            // sloped:
        g.drawLine(0,4,4,0);
        g.drawLine(right4,0,right,4);
        g.drawLine(0,bottom4,4,bottom);
        g.drawLine(right4,bottom,right,bottom4);
    }   // LifeCanvas.debugFrame()


    public void paint ( Graphics g )                    // redraws the whole visible playfield
    {
        synchronized ( engine ) {
            paintAll = false;
                // clear background:
            g.setColor(colors[0]);
            g.fillRect(1,1,myWidth-2,myHeight-2);
                // draw frame:
            g.setColor(Color.white);
            g.drawRect(0,0,myWidth-1,myHeight-1);
                // cell image cache:
            if ( imageCache == null ) createImageCache();
                // debugging:
            //debugFrame(g);
                // clipping:
            g.clipRect(1,1,myWidth-2,myHeight-2);
                // draw only live boxes within clipping region:
            LifeBox box;
            int boxSize = LifeBox.boxSize*cellSize;
                // loop over all live boxes:
            while ( (box = engine.getBox()) != null ) {
                if ( box.cellNumber > 0 ) {
                    int origX = (box.x0 - x0)*cellSize;
                    int origY = (box.y0 - y0)*cellSize;
                    if ( origX + boxSize > 1 && origX < myWidth-1 &&
                         origY + boxSize > 1 && origY < myHeight-1 ) {
                        int y1 = Math.max(0,y0-box.y0);
                        int y2 = Math.min(box.boxSize,y0+1+((myHeight-3)/cellSize)-box.y0);
                        int x1 = Math.max(0,x0-box.x0);
                        int x2 = Math.min(box.boxSize,x0+1+((myWidth-3)/cellSize)-box.x0);
                        int shift0 = box.wordSize - box.boxSize + x1;
                        int ix0 = origX + 1 + x1*cellSize;
                        int iy = origY + 1 + y1*cellSize;
                        for ( int y = y1; y < y2; y++, iy += cellSize ) {
                            int cells = box.cells[y] << shift0;
                            if ( LifeBox.maxAge > 1 ) {
                                for ( int x = x1, ix = ix0; cells != 0 && x < x2; x++, cells += cells, ix += cellSize )
                                    if ( cells < 0 )
                                        g.drawImage(imageCache[ box.ages[y][x] ],ix,iy,this);
                                }
                            else
                                for ( int x = x1, ix = ix0; cells != 0 && x < x2; x++, cells += cells, ix += cellSize )
                                    if ( cells < 0 )
                                        g.drawImage(imageCache[ (cells<0) ? 1 : 0 ],ix,iy,this);
                            }
                        }
                    }
                for ( int y = 0; y < box.boxSize; y++ )
                    box.draw[y] = 0;
                box.toDraw = false;
                }
                // draw color palette & stat strings:
            finishDraw(g,true);
            }
    }   // LifeCanvas.paint()


    public void clear ()                                // reinitializes the playfield
    {
        engine.finalize();
        engine = new LifeEngine();
        paintAll = true;
        x0 = y0 = 0;
    }   // LifeCanvas.clear()


    protected void repaintMe ( boolean force )
    {
        if ( !force && (engine.generation % repaintCycle) != 0 ) return;
        boolean paint = false;
        synchronized ( this ) {
            if ( !toPaint )
                toPaint = paint = true;
			}
        if ( paint ) repaint();
    }   // LifeCanvas.repaintMe()


    protected void repaintAll ()                        // deferred paint
    {
        paintAll = true;
        repaintMe(true);
    }   // LifeCanvas.repaintAll()


    public boolean mouseDown ( Event evt, int x, int y ) // mouse click event handler
    {
        int ix = x0 + (x-1) / cellSize;
        int iy = y0 + (y-1) / cellSize;
        engine.toggleCell(ix,iy);
        repaintMe(true);
        return true;
    }   // LifeCanvas.mouseDown()


    public void shift ( int horizontal, int vertical )  // shifts the visible window
    {
        int deltaX = horizontal*LifeBox.boxSize;
        int max = myWidth >> 1;
        while ( Math.abs(deltaX*cellSize) > max ) deltaX >>= 1;
        int deltaY = vertical*LifeBox.boxSize;
        max = myHeight >> 1;
        while ( Math.abs(deltaY*cellSize) > max ) deltaY >>= 1;
        x0 += deltaX;
        y0 += deltaY;
        repaintAll();
    }   // LifeCanvas.shiftLeft()


    public void zoomInOut ( int delta )                 // changes renderig zoom factor
    {
        if ( cellSize+delta <= 0 ) return;
        int middleX = x0 + myWidth / (cellSize+cellSize);
        int middleY = y0 + myHeight / (cellSize+cellSize);
        cellSize += delta;
        x0 = middleX - myWidth / (cellSize+cellSize);
        y0 = middleY - myHeight / (cellSize+cellSize);
        createImageCache();
        repaintAll();
    }   // LifeCanvas.zoomInOut()


    public boolean keyDown ( Event evt, int key )       // keyboard event handler
    {
        switch ( key ) {
            case Event.LEFT:
                shift(-1,0);
                return true;
            case Event.RIGHT:
                shift(1,0);
                return true;
            case Event.UP:
                shift(0,-1);
                return true;
            case Event.DOWN:
                shift(0,1);
                return true;
            }
        return false;
    }   // LifeCanvas.keyDown()


    public void nextGeneration ()                       // computes the next generation
    {
        engine.nextGeneration();
    }   // LifeCanvas.nextGeneration()


    public void setAges ( boolean ages )                // sets cells aging
    {
        engine.setMaxAge( ages ? maxAge : 1 );
    }   // LifeCanvas.setAges()

}   // LifeCanvas
