//******************************************************************************
//  LifeEngine.java:    LifeEngine class
//      (c) 1997 Josef Pelikan, KSVI MFF UK Praha
//          Josef.Pelikan@mff.cuni.cz, http://sun3.ms.mff.cuni.cz/~pepca/
//******************************************************************************
//  Abstract life engine: sparse playfield representation, computing of
//      next generation
//  Parameters: maximum cell age (1 .. disables cell aging)
//      rules (# of neighbours for cell survival/birth)
//  Playfield: 2D array of "boxes", each box stores max 30x30 cells (in int[] array),
//      non-mandatory age array (in byte[][] array), 32-bit hashing value (int), ..
//******************************************************************************

import java.awt.*;
import java.util.*;

//=======================================================================
//  LifeBox: cell box (max. 30x30 cells)
//      approx netto size (16x16): 9*4+3*64+72+(256)   = 300(556)
//                        (30x30): 9*4+3*120+128+(900) = 524(1424)
//=======================================================================

class LifeBox implements Cloneable
{
        // dimensions:
    public int x0, y0;              // origin (cell coordinate!)
    public final static int boxSize = 16;   // alternatives: 24, 30
    public final static int wordSize = 32;  // number of bits in cell word

        // playfield:
    protected int cells[];          // bit mask for max. 16x16 cells
    protected int active[];         // bit mask of active (=changed) cells
    protected int draw[];           // cells to be drawn
    protected int old[];            // old cell values (for next generation in-place)
    protected byte ages[][];        // (optional) array of cell ages: ages[row][column]
    public static int maxAge = 25;
                                    // maximum cell age (<2 disables cell aging)
    protected int cellNumber;       // total number of live cells
    protected int activeNumber;     // number of changed cells
    final static int inducted = 1024;   // box has been inducted from its neighbours
    protected boolean toDraw;       // cells to be drawn
    protected int activeBorder;     // borders with new cells
    protected int invalidBorder;    // invalid border cells (for next generation)

    protected int generation;       // last computed generation

    protected boolean prepared;     // flag for prepareOld()

        // box recycling:
    protected static LifeBox recycled = null;
    protected LifeBox next;         // linked list of recycled boxes
        // recycle statistics:
    protected static int recycles = 0;
    protected static int allocated = 0;


    protected void finalize ()      // recycle the box
    {
        next = recycled;
        recycled = this;
    }   // LifeBox.finalize()


    public static LifeBox newLifeBox ( int x, int y )
                                    // returns new LifeBox object (recycled)
    {
        if ( recycled == null ) {
            allocated++;
            return new LifeBox(x,y);
            }
        recycles++;
        LifeBox result = recycled;
        recycled = result.next;
        result.x0 = x; result.y0 = y;
        for ( int i = 0; i < boxSize; i++ )
            result.cells[i] = result.active[i] = result.draw[i] = 0;
        result.hashValue = result.activeBorder = result.cellNumber =
            result.activeNumber = result.generation = 0;
        result.toDraw = true;                   // new box has to be drawn
        if ( maxAge > 1 )
            if ( result.ages == null ) result.ages = new byte[boxSize][boxSize];
            else
                for ( y = 0; y < boxSize; y++ )
                    for ( x = 0; x < boxSize; x++ )
                        result.ages[y][x] = (byte)0;
        result.prepared = false;
        return result;
    }   // LifeBox.newLifeBox()


        // old cell masks:
    final static int highMask = 1 << (boxSize-1);
    final static int oldInLeft = highMask << 1;
    final static int oldLeft = oldInLeft << 1;
    final static int oldInside = oldLeft - 2;
    final static int oldRight = 1;
    final static int oldInRight = oldRight << 1;
        // border mask bits:
    final static int bUpperLeft  = 0x001;
    final static int bUpper      = 0x002;
    final static int bUpperRight = 0x004;
    final static int bLeft       = 0x008;
    final static int bRight      = 0x020;
    final static int bLowerLeft  = 0x040;
    final static int bLower      = 0x080;
    final static int bLowerRight = 0x100;
    final static int bAll = bUpperLeft | bUpper | bUpperRight | bLeft | bRight |
                            bLowerLeft | bLower | bLowerRight;
    
        // playfield hashing (depends on cell values):
    protected int hashValue;
    protected int hashPrev, hashPrevPrev;

    final static int rowHash  = 23; // cell row multiplier
    final static int x0Hash   = 13; // for superHash:
    final static int y0Hash   = 17;
    final static int hashHash = 19;

        // rules:
    protected static int surviveMask = (1<<2) | (1<<3);
                                    // bit mask for cell survivals (LSB .. 0 neighbours)
    protected static int bornMask = (1<<3);
                                    // bit mask for cell births

        // incremental next generation:
    protected static byte state[];  // automaton for next generation of K cells
                                    // 2^((K+2)*3) states: K+2 LSB .. rightmost cell column
                                    // byte state: 1-4 LSB .. lowest new cell,
                                    //             4 MSB .. change mask

    private final static int leftMask = 7 << 6;
    private final static int middleMask = 7 << 3;
    private final static int rightMask = 7;

    public void prepareAutomaton () // 4x3 cell matrix => 2 new cells
    {
        state = new byte[1<<12];    // 2^(4*3)
        byte one[] = prepareNeighbourhood();
                                    // one cell neighbourhood
        for ( int i = 0; i < (1<<12); i++ )
            state[i] = (byte)( (one[((i>>3) & leftMask) +
                                    ((i>>2) & middleMask) +
                                    ((i>>1) & rightMask)] << 1) +
                                one[((i>>2) & leftMask) +
                                    ((i>>1) & middleMask) +
                                    ( i     & rightMask)] );
    }   // LifeBox.prepareAutomaton()

    private byte[] prepareNeighbourhood ()
    {
        byte n[] = new byte[1<<9];
        for ( int i = 0; i < (1<<9); i++ ) {
            int neighbours = ( i     & 1) + ((i>>1) & 1) + ((i>>2) & 1) +
                             ((i>>3) & 1) +                ((i>>5) & 1) +
                             ((i>>6) & 1) + ((i>>7) & 1) + ((i>>8) & 1);
            if ( (i & (1<<4)) > 0 ) // survival?
                n[i] = (byte)( ((surviveMask >> neighbours) & 1) > 0 ?
                               0x01 : 0x10 );
            else                    // birth?
                n[i] = (byte)( ((bornMask >> neighbours) & 1) > 0 ?
                               0x11 : 0x00 );
            }
        return n;
    }   // LifeBox.prepareNeighbourhood()


        // LifeBox constructors:

    protected LifeBox ()                // simple constructor (for dummy objects)
    {
    }   // LifeBox.LifeBox()


    public LifeBox ( int x, int y )     // creates new empty cell box
    {
        x0 = x; y0 = y;
        cells = new int[boxSize];       // empty cell array
        active = new int[boxSize];
        draw = new int[boxSize];
        hashValue = activeBorder = cellNumber = activeNumber = generation = 0;
        toDraw = true;                  // new box has to be drawn
        if ( maxAge > 1 ) ages = new byte[boxSize][boxSize];
        prepared = false;
    }   // LifeBox.LifeBox(int,int)


    public Object clone ()
    {
        try {
            LifeBox newBox = (LifeBox)super.clone();
            newBox.x0 = x0;
            newBox.y0 = y0;
            newBox.cells = new int[boxSize];
            System.arraycopy(cells,0,newBox.cells,0,boxSize);
            newBox.active = new int[boxSize];
            System.arraycopy(active,0,newBox.active,0,boxSize);
            newBox.draw = new int[boxSize];
            System.arraycopy(draw,0,newBox.draw,0,boxSize);
            newBox.hashValue = hashValue;
            newBox.hashPrev = hashPrev;
            newBox.hashPrevPrev = hashPrevPrev;
            newBox.activeBorder = activeBorder;
            newBox.invalidBorder = invalidBorder;
            newBox.cellNumber = cellNumber;
            newBox.activeNumber = activeNumber;
            newBox.toDraw = toDraw;
            newBox.generation = generation;
            newBox.ages = ages;         // ages are shared!
            newBox.prepared = false;
            return newBox;
            }
        catch ( CloneNotSupportedException e ) {
            return null;
            }
    }   // LifeBox.clone()


    public void setMaxAge ()            // sets new cell aging (maxAge is the new value)
    {
        if ( maxAge > 1 ) {
            if ( ages == null )         // allocate new ages array
                ages = new byte[boxSize][boxSize];
            for ( int y = 0; y < boxSize; y++ ) {
                int M = cells[y];
                if ( M != 0 ) activeNumber = 1;
                for ( int x = boxSize, dM = 1; --x >= 0 ; dM += dM )
                    ages[y][x] = (byte)( (M & dM) != 0 ? 1 : 0 );
                }
            }
    }   // LifeBox.setMaxAge()


        // next generation:

    final static int[] number = { 0, 1, 1, 2 };

    public void prepareOld ( LifeEngine engine )
    {
        // unprocessed box should have invalidBorder==0 || activeNumber == 0
        // nextGeneration(): computed cells become invalidBorder = 0,
        //                   skipped cells have activeNumber == 0
        prepared = true;
            // old generation:
        if ( old == null )
            old = new int[boxSize+2];
        old[0] = old[boxSize+1] = 0;
        for ( int y = 0; y < boxSize; y++ )
            old[y+1] = cells[y] + cells[y];
        invalidBorder = bAll;
            // neighbours:
        for ( int mask = 1; mask <= activeBorder; mask += mask )
            if ( (activeBorder & mask) != 0 ) {
                int x = x0, y = y0;
                switch ( mask ) {
                    case bUpperRight:
                        x += boxSize;
                    case bUpper:
                        y -= boxSize;
                        break;
                    case bUpperLeft:
                        y -= boxSize;
                    case bLeft:
                        x -= boxSize;
                        break;
                    case bLowerRight:
                        y += boxSize;
                    case bRight:
                        x += boxSize;
                        break;
                    case bLowerLeft:
                        x -= boxSize;
                    case bLower:
                        y += boxSize;
                        break;
                    }
                LifeBox box = engine.field.get(x,y);
                if ( box == null )      // add new empty box:
                    (box = engine.newBox(x,y)).prepareOld(engine);
                box.activeNumber = inducted;        // recompute flag
                }
    }   // LifeBox.prepareOld()


    public void nextGeneration ( BoxHash field, int generation )
        // all boxes had to been prepared (by prepareOld)
        // cells with activeNumber == 0 can be skipped
    {
            // incremental new generation:
        hashPrevPrev = hashPrev; hashPrev = hashValue;
        hashValue = cellNumber = activeNumber = activeBorder = 0;
            // four edge-neighbours:
        LifeBox box;
            // upper:
        if ( (invalidBorder & bUpper) != 0 ) {
            box = field.get(x0,y0-boxSize);
            if ( box != null ) {        // data exchange with upper box:
                    // upper to this:
                if ( (box.invalidBorder & bLeft) == 0 )
                    invalidBorder &= bAll-bUpperLeft;
                if ( (box.invalidBorder & bRight) == 0 )
                    invalidBorder &= bAll-bUpperRight;
                old[0] |= box.old[boxSize];
                    // this to upper:
                if ( (invalidBorder & bLeft) == 0 )
                    box.invalidBorder &= bAll-bLowerLeft;
                if ( (invalidBorder & bRight) == 0 )
                    box.invalidBorder &= bAll-bLowerRight;
                box.old[boxSize+1] |= old[1];
                box.invalidBorder &= bAll-bLower;
                }
            invalidBorder -= bUpper;
            }
            // lower:
        if ( (invalidBorder & bLower) != 0 ) {
            box = field.get(x0,y0+boxSize);
            if ( box != null ) {        // data exchange with lower box:
                    // lower to this:
                if ( (box.invalidBorder & bLeft) == 0 )
                    invalidBorder &= bAll-bLowerLeft;
                if ( (box.invalidBorder & bRight) == 0 )
                    invalidBorder &= bAll-bLowerRight;
                old[boxSize+1] |= box.old[1];
                    // this to lower:
                if ( (invalidBorder & bLeft) == 0 )
                    box.invalidBorder &= bAll-bUpperLeft;
                if ( (invalidBorder & bRight) == 0 )
                    box.invalidBorder &= bAll-bUpperRight;
                box.old[0] |= old[boxSize];
                box.invalidBorder &= bAll-bUpper;
                }
            invalidBorder -= bLower;
            }
            // left:
        if ( (invalidBorder & bLeft) != 0 ) {
            box = field.get(x0-boxSize,y0);
            if ( box != null ) {        // data exchange with left box:
                    // left to this:
                if ( (box.invalidBorder & bUpper) == 0 )
                    invalidBorder &= bAll-bUpperLeft;
                if ( (box.invalidBorder & bLower) == 0 )
                    invalidBorder &= bAll-bLowerLeft;
                for ( int i = 0; i <= boxSize+1; i++ ) {
                    old[i] |= (box.old[i] & oldInRight) << boxSize;
                    // this to left:
                    box.old[i] |= (old[i] & oldInLeft) >> boxSize;
                    }
                box.invalidBorder &= bAll-bUpperRight-bRight-bLowerRight;
                }
            invalidBorder -= bLeft;
            }
            // right:
        if ( (invalidBorder & bRight) != 0 ) {
            box = field.get(x0+boxSize,y0);
            if ( box != null ) {        // data exchange with right box:
                    // right to this:
                if ( (box.invalidBorder & bUpper) == 0 )
                    invalidBorder &= bAll-bUpperRight;
                if ( (box.invalidBorder & bLower) == 0 )
                    invalidBorder &= bAll-bLowerRight;
                for ( int i = 0; i <= boxSize+1; i++ ) {
                    old[i] |= (box.old[i] & oldInLeft) >> boxSize;
                    // this to right:
                    box.old[i] |= (old[i] & oldInRight) << boxSize;
                    }
                box.invalidBorder &= bAll-bUpperLeft-bLeft-bLowerLeft;
                }
            invalidBorder -= bRight;
            }
            // upper left:
        if ( (invalidBorder & bUpperLeft) != 0 ) {
            box = field.get(x0-boxSize,y0-boxSize);
            if ( box != null ) {        // data exchange with upper left box:
                    // upper left to this:
                old[0] |= (box.old[boxSize] & oldInRight) << boxSize;
                    // this to upper left:
                if ( (box.invalidBorder & bLowerRight) != 0 ) {
                    box.old[boxSize+1] |= (old[1] & oldInLeft) >> boxSize;
                    box.invalidBorder -= bLowerRight;
                    }
                }
            invalidBorder -= bUpperLeft;
            }
            // upper right:
        if ( (invalidBorder & bUpperRight) != 0 ) {
            box = field.get(x0+boxSize,y0-boxSize);
            if ( box != null ) {        // data exchange with upper right box:
                    // upper right to this:
                old[0] |= (box.old[boxSize] & oldInLeft) >> boxSize;
                    // this to upper right:
                if ( (box.invalidBorder & bLowerLeft) != 0 ) {
                    box.old[boxSize+1] |= (old[1] & oldInRight) << boxSize;
                    box.invalidBorder -= bLowerLeft;
                    }
                }
            invalidBorder -= bUpperRight;
            }
            // lower left:
        if ( (invalidBorder & bLowerLeft) != 0 ) {
            box = field.get(x0-boxSize,y0+boxSize);
            if ( box != null ) {        // data exchange with lower left box:
                    // lower left to this:
                old[boxSize+1] |= (box.old[1] & oldInRight) << boxSize;
                    // this to lower left:
                if ( (box.invalidBorder & bUpperRight) != 0 ) {
                    box.old[0] |= (old[boxSize] & oldInLeft) >> boxSize;
                    box.invalidBorder -= bUpperRight;
                    }
                }
            invalidBorder -= bLowerLeft;
            }
            // lower right:
        if ( (invalidBorder & bLowerRight) != 0 ) {
            box = field.get(x0+boxSize,y0+boxSize);
            if ( box != null ) {        // data exchange with lower right box:
                    // lower right to this:
                old[boxSize+1] |= (box.old[1] & oldInLeft) >> boxSize;
                    // this to upper right:
                if ( (box.invalidBorder & bUpperLeft) != 0 ) {
                    box.old[0] |= (old[boxSize] & oldInRight) << boxSize;
                    box.invalidBorder -= bUpperLeft;
                    }
                }
            invalidBorder -= bLowerRight;
            }
        
        if ( state == null ) prepareAutomaton();
            // automata states for cell pairs:
        int s0, s1, s2, s3, s4, s5, s6, s7;
        s0 = ((old[0] <<  4) & 0xf0) | ( old[1]        & 0x0f);
        s1 = ((old[0] <<  2) & 0xf0) | ((old[1] >>  2) & 0x0f);
        s2 = ( old[0]        & 0xf0) | ((old[1] >>  4) & 0x0f);
        s3 = ((old[0] >>  2) & 0xf0) | ((old[1] >>  6) & 0x0f);
        s4 = ((old[0] >>  4) & 0xf0) | ((old[1] >>  8) & 0x0f);
        s5 = ((old[0] >>  6) & 0xf0) | ((old[1] >> 10) & 0x0f);
        s6 = ((old[0] >>  8) & 0xf0) | ((old[1] >> 12) & 0x0f);
        s7 = ((old[0] >> 10) & 0xf0) | ((old[1] >> 14) & 0x0f);
            // cell row loop:
        for ( int i = 0; i < boxSize; i++ ) {
            int o = old[i+2];
                // s0
            int st = state[s0 = ((s0 << 4) & 0xff0) | (o & 0x0f)];
            int c = (st & 3);
            cellNumber += number[st & 3];
            int M = (st & 0x30) >> 4;
            int num = number[(st >> 4) & 3];
                // s1
            st = state[s1 = ((s1 << 4) & 0xff0) | ((o >> 2) & 0x0f)];
            c += (st & 3) << 2;
            cellNumber += number[st & 3];
            M += (st & 0x30) >> 2;
            num += number[(st >> 4) & 3];
                // s2
            st = state[s2 = ((s2 << 4) & 0xff0) | ((o >> 4) & 0x0f)];
            c += (st & 3) << 4;
            cellNumber += number[st & 3];
            M += st & 0x30;
            num += number[(st >> 4) & 3];
                // s3
            st = state[s3 = ((s3 << 4) & 0xff0) | ((o >> 6) & 0x0f)];
            c += (st & 3) << 6;
            cellNumber += number[st & 3];
            M += (st & 0x30) << 2;
            num += number[(st >> 4) & 3];
                // s4
            st = state[s4 = ((s4 << 4) & 0xff0) | ((o >> 8) & 0x0f)];
            c += (st & 3) << 8;
            cellNumber += number[st & 3];
            M += (st & 0x30) << 4;
            num += number[(st >> 4) & 3];
                // s5
            st = state[s5 = ((s5 << 4) & 0xff0) | ((o >> 10) & 0x0f)];
            c += (st & 3) << 10;
            cellNumber += number[st & 3];
            M += (st & 0x30) << 6;
            num += number[(st >> 4) & 3];
                // s6
            st = state[s6 = ((s6 << 4) & 0xff0) | ((o >> 12) & 0x0f)];
            c += (st & 3) << 12;
            cellNumber += number[st & 3];
            M += (st & 0x30) << 8;
            num += number[(st >> 4) & 3];
                // s7
            st = state[s7 = ((s7 << 4) & 0xff0) | ((o >> 14) & 0x0f)];
            c += (st & 3) << 14;
            cellNumber += number[st & 3];
            M += (st & 0x30) << 10;
            num += number[(st >> 4) & 3];
                // final computing:
            activeNumber += num;
            cells[i] = c;                   // live cells
            active[i] = M;                  // changed cells
            hashValue = hashValue*rowHash + c;
                // active border:
            if ( i == 0 ) {                 // first row
                if ( M != 0 ) {
                    activeBorder = bUpper;
                    if ( (M & highMask) != 0 ) activeBorder |= bLeft + bUpperLeft;
                    if ( (M & 1) != 0 ) activeBorder |= bRight + bUpperRight;
                    }
                }
            else
                if ( i == boxSize-1 ) { // last row
                    if ( M != 0 ) {
                        activeBorder |= bLower;
                        if ( (M & highMask) != 0 ) activeBorder |= bLeft + bLowerLeft;
                        if ( (M & 1) != 0 ) activeBorder |= bRight + bLowerRight;
                        }
                    }
                else {
                    if ( (M & highMask) != 0 ) activeBorder |= bLeft;
                    if ( (M & 1) != 0 ) activeBorder |= bRight;
                    }
                // cell ages:
            if ( maxAge > 1 )
                for ( int x = boxSize, dM = 1; --x >= 0 ; dM += dM )
                    if ( (c & dM) != 0 ) {          // live cell
                        if ( ages[i][x] < maxAge &&
                             ages[i][x]++ != 0 ) {  // the cell is growing old
                            activeNumber++;
                            M |= dM;
                            }
                        }
                    else                            // dead cell
                        if ( (M & dM) != 0 ) ages[i][x] = (byte)0;
                // active & draw masks:
            draw[i] |= M;
            if ( M != 0 ) toDraw = true;
            }
        this.generation = generation;
    }   // LifeBox.nextGeneration


    public int rehash ()            // updates hash value, returns hash increment
    {
        int old = superHash();
        hashValue = cells[0];
        for ( int j = 1; j < boxSize; j++ )
            hashValue = hashValue*rowHash + cells[j];
        return( superHash() - old );
    }   // LifeBox.rehash


    public int superHash ()         // hashing function used by box owner
    {
        return( x0*x0Hash + y0*y0Hash + hashValue*hashHash );
    }   // LifeBox.superHash


    public void toggleCell ( int x, int y, LifeEngine engine )
        // returns hash increment
    {
        int mask = 1 << (boxSize-1-x);
            // draw mask:
        draw[y] |= mask;
        engine.toDraw = toDraw = true;
            // active border:
        if ( y == 0 ) {
            activeBorder |= bUpper;
            if ( x == 0 ) activeBorder |= bUpperLeft;
            if ( x == boxSize-1 ) activeBorder |= bUpperRight;
            }
        else
            if ( y == boxSize-1 ) {
                activeBorder |= bLower;
                if ( x == 0 ) activeBorder |= bLowerLeft;
                if ( x == boxSize-1 ) activeBorder |= bLowerRight;
                }
        if ( x == 0 ) activeBorder |= bLeft;
        else
            if ( x == boxSize-1 ) activeBorder |= bRight;
            // counters:
        boolean birth = ((cells[y] ^= mask) & mask) != 0;
        if ( birth ) {
            engine.cellNumber++;
            if ( cellNumber++ == 0 ) engine.nonemptyBoxes++;
            }
        else {
            engine.cellNumber--;
            if ( --cellNumber == 0 ) engine.nonemptyBoxes--;
            }
        if ( maxAge > 1 )
            ages[y][x] = (byte)(birth ? 1 : 0);
        if ( (active[y] & mask) == 0 ) {
            engine.activeNumber++;
            if ( activeNumber++ == 0 ) engine.activeBoxes++;
            }
        active[y] |= mask;
        engine.hashValue += rehash();
    }   // LifeBox.toggleCell()

}   // LifeBox


//=======================================================================
//  BoxHash: custom hashing array with fast access and simple iterator
//      maps [int,int] doubles to LifeBox references
//=======================================================================

class BoxHash implements Cloneable
{
        // LifeBox array:
    protected LifeBox array[];      // array size should be prime number
    protected int count;            // number of stored objects
    protected int removed;          // number of removed objects
    protected int threshold;        // overload threshold
    protected float loadFactor;     // maximum BoxHash load

        // rehashing #:
    protected int rehashNumber;

        // default values:
    protected final static int initArraySize = 31;
    protected final static float initLoadFactor = 0.4f;

        // removed objects:
    protected final static LifeBox empty = new LifeBox();


    public BoxHash ( int initSize, float initLoadFactor )   // complete info constructor
    {
        loadFactor = (initLoadFactor<0.1f) ? 0.1f : (initLoadFactor>0.9f) ? 0.9f : initLoadFactor;
        int size = (initSize/loadFactor < initArraySize) ?
                    initArraySize : (int)(initSize/loadFactor) | 1;
        threshold = (int)(size*loadFactor);
        array = new LifeBox[size];
        count = removed = rehashNumber = 0;
    }   // BoxHash.BoxHash(int,float)


    public BoxHash ( int initSize )                         // partial info constructor
    {
        this(initSize,initLoadFactor);
    }   // BoxHash.BoxHash(int)


    public BoxHash ()                                       // default constructor
    {
        this((int)(initArraySize*initLoadFactor));
    }   // BoxHash.BoxHash()


    public Object clone ()
    {
        try {
            BoxHash newHash = (BoxHash)super.clone();
            newHash.array = new LifeBox[array.length];
            System.arraycopy(array,0,newHash.array,0,array.length);
            newHash.count = count;
            newHash.removed = removed;
            newHash.threshold = threshold;
            newHash.loadFactor = loadFactor;
            return newHash;
            }
        catch ( CloneNotSupportedException e ) {
			return null;
			}
    }   // BoxHash.clone()


    private void resize ()                                  // resizes the hashed array
    {
        //System.out.println("resize: "+array.length+","+count+","+removed);
        int size = (array.length + array.length) | 1;
        LifeBox old[] = array;
        array = new LifeBox[size];
        count = removed = 0;
        threshold = (int)(size*loadFactor);
        for ( int i = 0; i < old.length; i++ )
            if ( old[i] != null && old[i] != empty ) put(old[i]);
        rehashNumber++;
    }   // BoxHash.resize()


    private void rehash ()                                  // compacts hashing array
    {
        //System.out.println("rehash: "+array.length+","+count+","+removed);
        removed = 0;
        for ( int i = 0; i < array.length; i++ )
            if ( array[i] != null && array[i] != empty &&   // try to rehash the valid box
                 put(array[i]) == null ) {                  // rehashed
                array[i] = empty;
                count--;
                }
        for ( int i = 0; i < array.length; i++ )
            if ( array[i] == empty ) array[i] = null;
        removed = 0;
        rehashNumber++;
    }   // BoxHash.rehash()


    public int size ()
    {
        return count;
    }   // BoxHash.size()


    public LifeBox put ( LifeBox box )                      // new assignment, returns old mapping
    {
        if ( box == null ) return null;
        empty.x0 = box.x0; empty.y0 = box.y0;
        int i = ((box.x0 ^ box.y0*16381) & 0x7fffffff) % array.length;
        while ( array[i] != null && (array[i].x0 != box.x0 || array[i].y0 != box.y0) )
            if ( ++i == array.length ) i = 0;
        LifeBox old = array[i];
        if ( old == empty ) {
            removed--;
            old = null;
            }
        array[i] = box;
        if ( old == null && ++count + removed > threshold )
            if ( removed > count ) rehash();
            else                   resize();
        return old;
    }   // BoxHash.put()


    public LifeBox get ( int x, int y )                     // returns mapped LifeBox or null
    {
        empty.x0 = x + 1;
        int i = ((x ^ y*16381) & 0x7fffffff) % array.length;
        while ( array[i] != null && (array[i].x0 != x || array[i].y0 != y) )
            if ( ++i == array.length ) i = 0;
        return array[i];
    }   // BoxHash.get()


    public LifeBox remove ( LifeBox box )                   // removes the given box
    {
        empty.x0 = box.x0 + 1;
        int i = ((box.x0 ^ box.y0*16381) & 0x7fffffff) % array.length;
        while ( array[i] != null && (array[i].x0 != box.x0 || array[i].y0 != box.y0) )
            if ( ++i == array.length ) i = 0;
        if ( array[i] == null ) return null;                // was not here
        array[i] = empty;
        removed++; count--;
        return box;
    }   // BoxHash.remove()


    public Point getIterator ( Point oldIterator )
    {
        if ( oldIterator == null )
            return new Point(0,rehashNumber);
        oldIterator.move(0,rehashNumber);
        return oldIterator;
    }   // BoxHash.getIterator()


    public LifeBox getNextBox ( Point iterator )
    {
        if ( iterator.y != rehashNumber )       // the array was re-arranged
            iterator.move(0,rehashNumber);      // some boxes can be passed twice!
        int i = iterator.x;
        while ( i < array.length && (array[i] == null || array[i] == empty) ) i++;
        iterator.x = i+1;
        if ( i < array.length ) return array[i];
        return null;
    }   // BoxHash.getNextBox()


    protected void finalize ()
    {
        for ( int i = 0; i < array.length; i++ ) {
            if ( array[i] != null && array[i] != empty )
                array[i].finalize();
            array[i] = null;
            }
    }   // BoxHash.finalize()

}   // BoxHash


//=======================================================================
//  LifeEngine: playfield representation, next generation computing
//=======================================================================

class LifeEngine implements Cloneable
{
        // playfield dimension:
    protected int fieldLeft;        // in cells
    protected int fieldTop;
    protected int fieldWidth;       // 0 for empty field
    protected int fieldHeight;

    public Rectangle fieldDimension ()
    {
        return new Rectangle( fieldLeft, fieldTop, fieldWidth, fieldHeight );
    }   // LifeEngine.fieldDimension


    public BoxHash field;           // hashed array of LifeBoxes (key: [x0,y0])
    protected final int initFieldSize = 12; // default BoxHash size


    public int hashValue;           // playfield hashing (depends on cell values)
    public int hashPrev;
    public int hashPrevPrev;

    public int cellNumber;          // total number of live cells
    public int activeNumber;        // number of changed cells
    public boolean toDraw;          // needs to be drawn?

    public int generation;          // current generation

        // LifeBox statistics:
    public int usedBoxes;
    public int activeBoxes;
    public int nonemptyBoxes;
    public int blinkerBoxes;        // boxes with period=2


    public LifeEngine ()
    {
        fieldWidth = fieldHeight = generation = blinkerBoxes =
            usedBoxes = activeBoxes = nonemptyBoxes = 0;
        toDraw = true;
        field = new BoxHash(initFieldSize);
    }   // LifeEngine.LifeEngine()


    public synchronized Object clone ()
    {
        try {
            LifeEngine newEngine = (LifeEngine)super.clone();
            newEngine.fieldLeft = fieldLeft;
            newEngine.fieldTop = fieldTop;
            newEngine.fieldWidth = fieldWidth;
            newEngine.fieldHeight = fieldHeight;
            newEngine.field = (BoxHash)field.clone();
            newEngine.hashValue = hashValue;
            newEngine.hashPrev = hashPrev;
            newEngine.hashPrevPrev = hashPrevPrev;
            newEngine.cellNumber = cellNumber;
            newEngine.activeNumber = activeNumber;
            newEngine.toDraw = toDraw;
            newEngine.generation = generation;
            newEngine.usedBoxes = usedBoxes;
            newEngine.activeBoxes = activeBoxes;
            newEngine.nonemptyBoxes = nonemptyBoxes;
            newEngine.blinkerBoxes = blinkerBoxes;
            return newEngine;
            }
        catch ( CloneNotSupportedException e ) {
			return null;
			}
    }   // LifeEngine.clone()


    public synchronized void nextGeneration ()      // computes the next generation (in-place)
    {
        generation++;
            // box statistics:
        hashPrevPrev = hashPrev; hashPrev = hashValue;
        cellNumber = activeNumber = hashValue = usedBoxes = nonemptyBoxes =
            activeBoxes = blinkerBoxes = 0;

            // prepare all relevant boxes:
        LifeBox box;
        Point iterator;
        for ( iterator = field.getIterator(null);
              (box = field.getNextBox(iterator)) != null; )
            if ( !box.prepared ) box.prepareOld(this);
            // box advance loop:
        for ( iterator = field.getIterator(iterator);
              (box = field.getNextBox(iterator)) != null; ) // compute one box:
            if ( box.activeNumber + box.cellNumber == 0 )   // the unactive box
                if ( box.toDraw ) {                         // wait until the box is updated on the screen
                    box.prepared = false;               // for the next generation
                    usedBoxes++;
                    }
                else
                    field.remove(box).finalize();
            else {
                box.prepared = false;                   // for the next generation
                usedBoxes++;
                if ( box.activeNumber != 0 ) {          // active box => recompute
                    box.nextGeneration(field,generation);
                    activeNumber += box.activeNumber;
                    if ( box.activeNumber != 0 ) activeBoxes++;
                    if ( box.toDraw ) toDraw = true;
                    hashValue += box.superHash();
                    if ( box.hashValue == box.hashPrevPrev && box.hashValue != box.hashPrev )
                        blinkerBoxes++;
                    }
                    // statistics:
                cellNumber += box.cellNumber;
                if ( box.cellNumber != 0 ) nonemptyBoxes++;
                }
    }   // LifeEngine.nextGeneration()


    protected LifeBox newBox ( int x, int y )       // creates new empty box
    {
        LifeBox box = LifeBox.newLifeBox(x,y);
        field.put(box);             // insert into new field
        if ( fieldWidth == 0 || x < fieldLeft ) {
            if ( fieldWidth != 0 ) fieldWidth += fieldLeft-x;
            fieldLeft = x;
            }
        if ( x >= fieldLeft + fieldWidth ) fieldWidth = x - fieldLeft + box.boxSize;
        if ( fieldHeight == 0 || y < fieldTop ) {
            if ( fieldHeight != 0 ) fieldHeight += fieldTop-y;
            fieldTop = y;
            }
        if ( y >= fieldTop + fieldHeight ) fieldHeight = y - fieldTop + box.boxSize;
        return box;
    }   // LifeEngine.newBox()


    private Point boxEnum;          // used by getBox()

    public LifeBox getBox ()                        // returns next box or null at the end of list
    {
        if ( boxEnum == null )      // first box
            boxEnum = field.getIterator(null);
        LifeBox next = field.getNextBox(boxEnum);
        if ( next != null ) return next;
        boxEnum = null;
        return null;
    }   // LifeEngine.getBox()


    public synchronized void toggleCell ( int x, int y )    // toggle the given cell
    {
        int dx = x % LifeBox.boxSize;           // relative x-coordinate
        if ( dx < 0 ) dx += LifeBox.boxSize;
        int dy = y % LifeBox.boxSize;           // relative y-coordinate
        if ( dy < 0 ) dy += LifeBox.boxSize;
        x -= dx; y -= dy;
        LifeBox box = field.get(x,y);
        if ( box == null ) {                    // insert new box
            box = newBox(x,y);
            usedBoxes++;
            }
            // toggle the cell:
        box.toggleCell(dx,dy,this);
    }   // LifeEngine.toggleCell()


    public synchronized void setMaxAge ( int maxAge )       // sets new cell aging
    {
        LifeBox.maxAge = maxAge;
        LifeBox box;
        for ( Point iterator = field.getIterator(null);
              (box = field.getNextBox(iterator)) != null; )
            box.setMaxAge();
    }   // LifeEngine.setMaxAge()


    protected void finalize ()
    {
        field.finalize();
        field = null;
    }   // LifeEngine.finalize()

}   // LifeEngine
