/// Wavelet transform in memory.
class Lifting {

  /**
   *  Computes forward TS-transform in memory.
   *   @param  n0 Size of the source segment (<code>s0</code>).
   *   @param  a Global array holding all segments.
   *   @param  s0 Start of the source segment.
   *   @param  s1 Start of the smooth destination segment.
   *   @param  d1 Start of the wavelet destination segment.
   */
  public static void liftMemory ( int n0, int[] a, int s0, int s1, int d1 ) {
    int n1 = n0 >> 1;
    int ps0, ps1, pd1;
    int i;
      // split:
    ps0 = s0;
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
      a[ps1++] = a[ps0++];
      a[pd1++] = a[ps0++];
    }
      // S-transform:
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
      a[pd1]   -= a[ps1];
      a[ps1++] += (a[pd1++] >> 1);
    }
      // predict:
    ps1 = s1; pd1 = d1;
    a[ps1+n1] = a[ps1+n1-1];           // boundary
    a[ps1-1]  = a[ps1];
    for ( i = 0; i++ < n1; ) {
      a[pd1++] += ((a[ps1-1] - a[ps1+1] + 2) >> 2);
      ps1++;
    }
  }

  /**
   *  Computes backward TS-transform in memory.
   *   @param  n0 Size of the destination segment (<code>s0</code>).
   *   @param  a Global array holding all segments.
   *   @param  s0 Start of the destination segment.
   *   @param  s1 Start of the smooth source segment.
   *   @param  d1 Start of the wavelet source segment.
   */
  public static void unliftMemory ( int n0, int[] a, int s0, int s1, int d1 ) {
    int n1 = n0 >> 1;
    int ps0, ps1, pd1;
    int i;
      // predict*:
    ps1 = s1; pd1 = d1;
    a[ps1+n1] = a[ps1+n1-1];           // boundary
    a[ps1-1]  = a[ps1];
    for ( i = 0; i++ < n1; ) {
      a[pd1++] -= ((a[ps1-1] - a[ps1+1] + 2) >> 2);
      ps1++;
    }
      // S-transform*:
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
      a[ps1]   -= (a[pd1] >> 1);
      a[pd1++] += a[ps1++];
    }
      // merge:
    ps0 = s0;
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
      a[ps0++] = a[ps1++];
      a[ps0++] = a[pd1++];
    }
  }

}

/*
  Wavelet (T-S) transform in memory (int32[4M-2K] arrays):
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,I5.0.1)        13.80s  100%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH)               13.98s  101%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.3.1IBM)     15.64s  113%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.4.1jrockit) 16.45s  119%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J5.00.3802)    18.48s  134%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.4.2bHS)     18.83s  136%
*/
