/**
 * Entity which can be positioned on the field.
 * Locator user, intellect and strength bearer.
 * @see     Locator
 */
interface Entity {

	/**
	 * Returns exact (private) position of the entity.
	 * @param  result Pre-allocated array to hold the result (optional, may be null).
	 * @return array holding two items: [x,y]
	 */
	int[] exactPosition ( int[] result );

	/** Sets exact (private) position of the entity. */
	void setExactPosition ( int x, int y );

	/** Intellect of the entity (ranges between 0.0f and 1.0f). */
	float intellect ();

	/** (Physical) strength of the entity (ranges between 0.0f and 1.0f). */
	float strength ();

}

/**
 * Object which can be simulated within time.
 */
interface Simulated {

	/**
	 * Starts simulation of one cycle (from time to time+delta).
	 * @see     #finishCycle
	 */
	void startCycle ( float time, float delta );

	/**
	 * Finishes simulation of one cycle (started by startCycle).
	 * @return  true if the entity weathers the period.
	 * @see     #startCycle
	 */
	boolean finishCycle ();

	/**
	 * Returns directions which the entity wants to go to in the actual
	 * simulation cycle. If Direction.STAY is not included, it will
	 * be appended automatically at the end of the list.
	 * @param   result Pre-allocated array to hold the result (optional, may be null).
	 *                 Size of the result array should be at least 10.
	 * @return  Sequence of wish-directions ordered by preference (descending order).
	 *          Direction.VOID is used as terminator.
	 * @see     Direction
	 */
	byte[] wishes ( byte[] result );

}

/** Entity which can be simulated both on the field and within time. */
interface SimulatedEntity extends Entity, Simulated {
}
